# -*- coding: utf-8 -*-
"""
ALPHA VERSION, PRELIMINARY, MANY NOT-YET-IMPLEMENTED FUNCTIONS AND FEATURES, CONTACT ACCES I/O PRODUCTS FOR SUPPORT IF NEEDED
Provide Python access via ctypes to ACCES's PCI, PCI Express, ISA, PC/104, and related DAQ devices installed using the AIOWDM.dll (Windows) library.

Consult https://accesio.com/MANUALS/Software%20Reference.pdf for additional information about the underlying functionality exposed by the API.
"""
from typing import *
from threading import Thread
from ctypes import *
AIOWDM = cdll.LoadLibrary("aiowdm")


def GetNumCards():int
    """Return the number of detected, installed, AIOWDM-using cards

    Returns:
        int: numCards, the number of AIOWDM.sys-detected cards in the system; 0 <= cardNum < numCards
    """
    return AIOWDM.GetNumCards()


def QueryCardInfo(cardNum):
    """
    Return status, PID, {baseaddress}, and card model name.

    It is virtually never necessary to use the baseaddress value due to the new "Relative" functions in the API: instead use RelOutPort*() and RelInPort*().
    """
    PID = c_ulong(0)
    base = c_ulong(0)
    namesize = c_ulong(50)
    nameBuf = create_string_buffer(50)
    status = AIOWDM.QueryCardInfo(cardNum, byref(
        PID), byref(base), byref(namesize), nameBuf)
    return status, PID.value, base.value, nameBuf.value.strip().decode()


def WaitForIRQ(cardNum):
    """
    This function deadlocks the calling thread until an IRQ occurs or the request is aborted.

    It's intended for multi-threaded IRQ handling, where the calling application spawns a separate thread to service IRQs quickly.
    Returns:
        nonzero if it returned because an IRQ occurred 
        Zero if it returned because the wait was aborted or there was an error. If the wait was aborted, GetLastError() will return ERROR_OPERATION_ABORTED; If someone is already waiting for an IRQ on that card when you call WaitForIRQ() it will return ERROR_INVALID_FUNCTION. (Other errors will return other error codes.)    
    """
    return AIOWDM.WaitForIRQ(cardNum)


def AbortRequest(cardNum):
    """Aborts WaitForIRQ and COSWaitForIRQ allowing the thread to terminate"""
    return AIOWDM.AbortRequest(cardNum)


def CloseCard(cardNum):
    """Aborts pending Waits and waits for completion.  Should be reliably called before your process ends"""
    return AIOWDM.CloseCard(cardNum)


def COSWaitForIRQ(cardNum, PPIs, data):
    """Blocks the calling thread until an IRQ occurs, and returns data read from the device in the ISR"""
    dataBuf = (c_ubyte*PPIs*3)()
    status = AIOWDM.COSWaitForIRQ(cardNum, PPIs, dataBuf)
    return status, dataBuf


def RelInPortB(cardNum, offset):
    """Reads a byte register"""
    return AIOWDM.RelInPortB(cardNum, offset)


def RelInPort(cardNum, offset):
    """Reads a UInt16 register"""
    return AIOWDM.RelInPort(cardNum, offset)


def RelInPortL(cardNum, offset):
    """Reads a UInt32 register"""
    return AIOWDM.RelInPortL(cardNum, offset)


def RelOutPortB(cardNum, offset, data):
    """Writes a byte register"""
    return AIOWDM.RelOutPortB(cardNum, offset, c_ubyte(data))


def RelOutPort(cardNum, offset, data):
    """Writes a UInt16 register"""
    return AIOWDM.RelOutPort(cardNum, offset, c_ushort(data))


def RelOutPortL(cardNum, offset, data):
    """Writes a UInt32 register"""
    return AIOWDM.RelOutPortL(cardNum, offset, c_ulong(data))


def InPortB(address):
    """Reads a byte register; deprecated, use RelInPortB instead when viable"""
    return AIOWDM.InPortB(address)


def InPort(address):
    """Reads a UInt16 register; deprecated, use RelInPort instead when viable"""
    return AIOWDM.InPort(address)


def InPortL(address):
    """Reads a UInt32 register; deprecated, use RelInPortL instead when viable"""
    return AIOWDM.InPortL(address)


def OutPortB(address, data):
    """Writes a byte register; deprecated, use RelOutPortB instead when viable"""
    return AIOWDM.OutPortB(address, c_ubyte(data))


def OutPort(address, data):
    """Writes a UInt16 register; deprecated, use RelOutPort instead when viable"""
    return AIOWDM.OutPort(address, c_ushort(data))


def OutPortL(address, data):
    """Writes a UInt32 register; deprecated, use RelOutPortL instead when viable"""
    return AIOWDM.OutPortL(address, c_ulong(data))


def displayBoardInfo(index):
    # we don't generally need the base address in this, the "RelInPort()" era
    status, PID, base, name = QueryCardInfo(index)
    print(index, PID, name)
    print("Card found at cardNum {0} has deviceID 0x{1:04X} = {2} [addr=0x{3:04X}]".format(
        index, PID, name, base))


if __name__ == "__main__":  # main code section

    class IRQWorker(Thread):
        def __init__(self):
            Thread.__init__(self)
            self.count = 0
            self.terminated = False

        def run(self):
            while not self.terminated:
                status = WaitForIRQ(0)
                if status != 0:
                    print("IRQ ! ", self.count)
                    self.count = self.count+1
                else:
                    print("terminated, status = ", status)
                    if status != 995:
                        print("error was ", GetLastError())

    print("AIOWDM Python introductory sample using ctypes\n")

    cardCount = GetNumCards()
    if cardCount == 0:
        print("No valid AIOWDM devices were detected.\nPlease confirm your device is installed\nand the drivers are installed correctly (check Device Manager).")
        exit()

    boardsFound = 0
    for cardNum in range(0, cardCount):
        boardsFound = boardsFound + 1
        displayBoardInfo(cardNum)
    print("\nAIOWDM Boards found:", boardsFound)

    if False:
        worker = IRQWorker()
        worker.start()
        RelInPortB(0, 2)  # enable IRQs

        input("Waiting for IRQs to occur. Press ENTER to continue.")

        print("Total IRQs detected:", worker.count)
        worker.terminated = True
        CloseCard(0)

    print("\nDone.")
